/* securitytoken.c - security token
 * 
 * Copyright (C) 2006 Ray Strode <rstrode@redhat.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 *
 * TODO:     - doing this per project is a bad idea i think.
 *             We should probably make this a system service 
 *             and use dbus.
 *
 *           - We hardcode a driver right now.  We should probably
 *             look up the default list and go from there. 
 */
#define SC_SECURITY_TOKEN_ENABLE_INTERNAL_API
#include "securitytoken.h"

#include <errno.h>
#include <string.h>
#include <unistd.h>

#include <glib.h>
#include <glib/gi18n.h>

#include <cert.h>
#include <nss.h>
#include <pk11func.h>
#include <prerror.h>
#include <secmod.h>
#include <secerr.h>

#if defined (SC_SECURITY_TOKEN_ENABLE_TEST) || defined (SC_SECURITY_TOKEN_MONITOR_ENABLE_TEST)
#define sc_debug(format, args...) g_printerr (format "\n", ##args)
#define sc_warning(format, args...) g_printerr (format "\n", ##args)
#else
#define sc_debug(format, args...) 
#define sc_warning(format, args...) 
#endif

struct _ScSecurityTokenPrivate {
	SECMODModule *module;
	ScSecurityTokenState state;

	CK_SLOT_ID slot_id;
	gint slot_series;

	PK11SlotInfo *slot;
	gchar *name;

	CERTCertificate *signing_certificate;
	CERTCertificate *encryption_certificate;
};

static void sc_security_token_finalize (GObject *object);
static void sc_security_token_class_install_signals (ScSecurityTokenClass *token_class);
static void sc_security_token_class_install_properties (ScSecurityTokenClass *token_class);
static void sc_security_token_set_property (GObject       *object,
					    guint          prop_id,
					    const GValue  *value,
					    GParamSpec    *pspec);
static void sc_security_token_get_property (GObject     *object,
					    guint        prop_id,
					    GValue      *value,
					    GParamSpec  *pspec);
static void sc_security_token_set_name (ScSecurityToken *token, const gchar *name);
static void sc_security_token_set_slot_id (ScSecurityToken *token, 
					   gint             slot_id);
static void sc_security_token_set_slot_series (ScSecurityToken *token, 
					       gint             slot_series);
static void sc_security_token_set_module (ScSecurityToken *token,
					  SECMODModule    *module);

static PK11SlotInfo *sc_security_token_find_slot_from_id (ScSecurityToken *token,
							  gint slot_id);

static PK11SlotInfo *sc_security_token_find_slot_from_token_name (ScSecurityToken *token,
								  const gchar     *token_name);
static gboolean sc_security_token_fetch_certificates (ScSecurityToken *token);


#ifndef SC_SECURITY_TOKEN_DEFAULT_SLOT_ID 
#define SC_SECURITY_TOKEN_DEFAULT_SLOT_ID ((gulong) -1)
#endif

#ifndef SC_SECURITY_TOKEN_DEFAULT_SLOT_SERIES
#define SC_SECURITY_TOKEN_DEFAULT_SLOT_SERIES -1
#endif

enum {
	PROP_0 = 0,
	PROP_NAME,
	PROP_SLOT_ID,
	PROP_SLOT_SERIES,
	PROP_MODULE,
	NUMBER_OF_PROPERTIES
};

enum {
	INSERTED,
	REMOVED,
	NUMBER_OF_SIGNALS
};

static guint sc_security_token_signals[NUMBER_OF_SIGNALS];

G_DEFINE_TYPE (ScSecurityToken, sc_security_token, G_TYPE_OBJECT);

static void
sc_security_token_class_init (ScSecurityTokenClass *token_class)
{
    GObjectClass *gobject_class;

    gobject_class = G_OBJECT_CLASS (token_class);

    gobject_class->finalize = sc_security_token_finalize;

    sc_security_token_class_install_signals (token_class);
    sc_security_token_class_install_properties (token_class);

    g_type_class_add_private (token_class,
			      sizeof (ScSecurityTokenPrivate));
}

static void
sc_security_token_class_install_signals (ScSecurityTokenClass *token_class)
{
    GObjectClass *object_class;

    object_class = G_OBJECT_CLASS (token_class);

    sc_security_token_signals[INSERTED] =
	    g_signal_new ("inserted",
			  G_OBJECT_CLASS_TYPE (object_class),
			  G_SIGNAL_RUN_LAST,
			  G_STRUCT_OFFSET (ScSecurityTokenClass,
					   inserted), 
			  NULL, NULL, g_cclosure_marshal_VOID__VOID, 
			  G_TYPE_NONE, 0);
    token_class->inserted = NULL;

    sc_security_token_signals[REMOVED] =
	    g_signal_new ("removed",
			  G_OBJECT_CLASS_TYPE (object_class),
			  G_SIGNAL_RUN_LAST,
			  G_STRUCT_OFFSET (ScSecurityTokenClass,
					   removed), 
			  NULL, NULL, g_cclosure_marshal_VOID__VOID, 
			  G_TYPE_NONE, 0);
    token_class->removed = NULL;
}

static void
sc_security_token_class_install_properties (ScSecurityTokenClass *token_class)
{
    GObjectClass *object_class;
    GParamSpec *param_spec;

    object_class = G_OBJECT_CLASS (token_class);
    object_class->set_property = sc_security_token_set_property;
    object_class->get_property = sc_security_token_get_property;

    param_spec = g_param_spec_ulong ("slot-id", _("Slot ID"),
				   _("The slot the token is in"),
				   1, G_MAXULONG,
				   SC_SECURITY_TOKEN_DEFAULT_SLOT_ID,
				   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
    g_object_class_install_property (object_class, PROP_SLOT_ID, param_spec);

    param_spec = g_param_spec_int ("slot-series", _("Slot Series"),
				   _("per-slot token identifier"),
				   -1, G_MAXINT,
				   SC_SECURITY_TOKEN_DEFAULT_SLOT_SERIES,
				   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
    g_object_class_install_property (object_class, PROP_SLOT_SERIES, param_spec);

    param_spec = g_param_spec_string ("name", _("name"),
				      _("name"), NULL,
				      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
    g_object_class_install_property (object_class, PROP_NAME, param_spec);

    param_spec = g_param_spec_pointer ("module", _("Module"),
				       _("security token driver"),
				       G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY);
    g_object_class_install_property (object_class, PROP_MODULE, param_spec);
}

static void 
sc_security_token_set_property (GObject       *object,
				guint          prop_id,
				const GValue  *value,
				GParamSpec    *pspec)
{
    ScSecurityToken *token = SC_SECURITY_TOKEN (object);

    switch (prop_id)
    {
	    case PROP_NAME:
		    sc_security_token_set_name (token, g_value_get_string (value));
		    break;

	    case PROP_SLOT_ID:
		    sc_security_token_set_slot_id (token, 
						   g_value_get_ulong (value));
		    break;

	    case PROP_SLOT_SERIES:
		    sc_security_token_set_slot_series (token, 
						       g_value_get_int (value));
		    break;

	    case PROP_MODULE:
		    sc_security_token_set_module (token, 
						  (SECMODModule *)
						  g_value_get_pointer (value));
		    break;

	    default:
		    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

CK_SLOT_ID
sc_security_token_get_slot_id (ScSecurityToken *token)
{
    return token->priv->slot_id;
}

ScSecurityTokenState 
sc_security_token_get_state (ScSecurityToken *token)
{
    return token->priv->state;
}

gchar *
sc_security_token_get_name (ScSecurityToken *token)
{
    return g_strdup (token->priv->name);
}

gboolean
sc_security_token_is_login_token (ScSecurityToken *token)
{
    const gchar *login_token_name;
    login_token_name = g_getenv ("PKCS11_LOGIN_TOKEN_NAME");

    if ((login_token_name == NULL) || (token->priv->name == NULL))
	    return FALSE;

    if (strcmp (token->priv->name, login_token_name) == 0)
	    return TRUE;

    return FALSE;
}

static void 
sc_security_token_get_property (GObject    *object,
				guint        prop_id,
				GValue      *value,
				GParamSpec  *pspec)
{
    ScSecurityToken *token = SC_SECURITY_TOKEN (object);

    switch (prop_id)
    {
	    case PROP_NAME:
		    g_value_take_string (value, 
					 sc_security_token_get_name (token));
		    break;

	    case PROP_SLOT_ID:
		    g_value_set_ulong (value, 
				       (gulong) sc_security_token_get_slot_id (token));
		    break;

	    case PROP_SLOT_SERIES:
		    g_value_set_int (value, 
				     sc_security_token_get_slot_series (token));
		    break;

	    default:
		    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
sc_security_token_set_name (ScSecurityToken *token,
			    const gchar     *name)
{
    if (name == NULL)
	    return;

    if ((token->priv->name == NULL) ||
	(strcmp (token->priv->name, name) != 0)) {
	    g_free (token->priv->name);
	    token->priv->name = g_strdup (name);

	    if (token->priv->slot == NULL) {
		    token->priv->slot = sc_security_token_find_slot_from_token_name (token,
										     token->priv->name);

		    if (token->priv->slot != NULL) {
			    gint slot_id, slot_series;

			    slot_id = PK11_GetSlotID (token->priv->slot);
			    if (slot_id != token->priv->slot_id)
				    sc_security_token_set_slot_id (token, slot_id);

			    slot_series = PK11_GetSlotSeries (token->priv->slot);
			    if (slot_series != token->priv->slot_series)
				    sc_security_token_set_slot_series (token, slot_series);

			    _sc_security_token_set_state (token, SC_SECURITY_TOKEN_STATE_INSERTED);
		    } else {
			    _sc_security_token_set_state (token, SC_SECURITY_TOKEN_STATE_REMOVED);
		    }
	    }


	    g_object_notify (G_OBJECT (token), "name");
    }
}

static void
sc_security_token_set_slot_id (ScSecurityToken *token,
			       gint             slot_id)
{
    if (token->priv->slot_id != slot_id)
    {
	    token->priv->slot_id = slot_id;

	    if (token->priv->slot == NULL) {
		    token->priv->slot = sc_security_token_find_slot_from_id (token,
									     token->priv->slot_id);

		    if (token->priv->slot != NULL) {
			    const gchar *token_name;

			    token_name = PK11_GetTokenName (token->priv->slot);
			    if ((token->priv->name == NULL) || 
				((token_name != NULL) &&
				(strcmp (token_name, token->priv->name) != 0)))
				    sc_security_token_set_name (token, token_name);

			    _sc_security_token_set_state (token, SC_SECURITY_TOKEN_STATE_INSERTED);
		    } else {
			    _sc_security_token_set_state (token, SC_SECURITY_TOKEN_STATE_REMOVED);
		    }
	    }

	    g_object_notify (G_OBJECT (token), "slot-id");
    }
}

static void
sc_security_token_set_slot_series (ScSecurityToken *token,
				   gint             slot_series)
{
    if (token->priv->slot_series != slot_series)
    {
	    token->priv->slot_series = slot_series;
	    g_object_notify (G_OBJECT (token), "slot-series");
    }
}

static void
sc_security_token_set_module (ScSecurityToken *token,
			      SECMODModule    *module)
{
    gboolean should_notify;

    if (token->priv->module != module)
	    should_notify = TRUE;
    else
	    should_notify = FALSE;

    if (token->priv->module != NULL) {
	    SECMOD_DestroyModule (token->priv->module);
	    token->priv->module = NULL;
    }

    if (module != NULL)
	    token->priv->module = SECMOD_ReferenceModule (module);

    if (should_notify)
	    g_object_notify (G_OBJECT (token), "module");
}

gint 
sc_security_token_get_slot_series (ScSecurityToken *token)
{
    return token->priv->slot_series;
}

static void
sc_security_token_init (ScSecurityToken *token)
{

    sc_debug ("initializing security token ");

    token->priv = G_TYPE_INSTANCE_GET_PRIVATE (token,
					       SC_TYPE_SECURITY_TOKEN,
					       ScSecurityTokenPrivate);

    if (token->priv->slot != NULL)
	    token->priv->name = g_strdup (PK11_GetTokenName (token->priv->slot));
}

static void sc_security_token_finalize (GObject *object)
{
    ScSecurityToken *token;
    GObjectClass *gobject_class;

    token = SC_SECURITY_TOKEN (object);

    g_free (token->priv->name);

    sc_security_token_set_module (token, NULL);

    gobject_class =
	    G_OBJECT_CLASS (sc_security_token_parent_class);

    gobject_class->finalize (object);
}

GQuark sc_security_token_error_quark (void)
{
    static GQuark error_quark = 0;

    if (error_quark == 0)
	    error_quark = g_quark_from_static_string ("sc-security-token-error-quark");

    return error_quark;
}

ScSecurityToken *
_sc_security_token_new (SECMODModule *module,
			CK_SLOT_ID    slot_id,
			gint          slot_series)
{
    ScSecurityToken *token;

    g_return_val_if_fail (module != NULL, NULL);
    g_return_val_if_fail (slot_id >= 1, NULL);
    g_return_val_if_fail (slot_series > 0, NULL);
    g_return_val_if_fail (sizeof (gulong) == sizeof (slot_id), NULL);

    token = SC_SECURITY_TOKEN (g_object_new (SC_TYPE_SECURITY_TOKEN,
					     "module", module,
					     "slot-id", (gulong) slot_id,
					     "slot-series", slot_series,
					     NULL));
    return token;
}

ScSecurityToken *
_sc_security_token_new_from_name (SECMODModule *module,
				  const gchar  *name)
{
    ScSecurityToken *token;

    g_return_val_if_fail (module != NULL, NULL);
    g_return_val_if_fail (name != NULL, NULL);

    token = SC_SECURITY_TOKEN (g_object_new (SC_TYPE_SECURITY_TOKEN,
					     "module", module,
					     "name", name,
					     NULL));
    return token;
}

void 
_sc_security_token_set_state (ScSecurityToken      *token,
			      ScSecurityTokenState  state)
{
    /* sc_security_token_fetch_certificates (token); */
    if (token->priv->state != state)
    {
	    token->priv->state = state;

	    if (state == SC_SECURITY_TOKEN_STATE_INSERTED) {
		    g_signal_emit (token, sc_security_token_signals[INSERTED], 0);
	    } else if (state == SC_SECURITY_TOKEN_STATE_REMOVED)
		    g_signal_emit (token, sc_security_token_signals[REMOVED], 0);
	    else
		    g_assert_not_reached ();
    }
}

/* So we could conceivably make the closure data a pointer to the token
 * or something similiar and then emit signals when we want passwords,
 * but it's probably easier to just get the password up front and use
 * it.  So we just take the passed in g_malloc'd (well probably, who knows)
 * and strdup it using NSPR's memory allocation routines.
 */
static char *
sc_security_token_password_handler (PK11SlotInfo *slot, 
				    PRBool        is_retrying, 
				    const gchar  *password)
{
    if (is_retrying)
	    return NULL;

    return password != NULL? PL_strdup (password): NULL;
}

gboolean
sc_security_token_unlock (ScSecurityToken *token,
			  const gchar     *password)
{   
    SECStatus status;

    PK11_SetPasswordFunc ((PK11PasswordFunc) sc_security_token_password_handler);

    /* we pass PR_TRUE to load certificates
     */
    status = PK11_Authenticate (token->priv->slot, PR_TRUE, (gpointer) password);

    if (status != SECSuccess) {
	    sc_debug ("could not unlock token - %d", status);
	    return FALSE;
    }
    return TRUE;
}

static PK11SlotInfo *
sc_security_token_find_slot_from_token_name (ScSecurityToken *token,
					     const gchar     *token_name)
{
    int i;

    for (i = 0; i < token->priv->module->slotCount; i++) {
	    const gchar *slot_token_name;

	    slot_token_name = PK11_GetTokenName (token->priv->module->slots[i]);

	    if ((slot_token_name != NULL) &&
		(strcmp (slot_token_name, token_name) == 0))
		    return token->priv->module->slots[i];
    }

    return NULL;
}

static PK11SlotInfo *
sc_security_token_find_slot_from_id (ScSecurityToken *token,
                                     gint slot_id)
{
    int i;

    for (i = 0; i < token->priv->module->slotCount; i++)
	    if (PK11_GetSlotID (token->priv->module->slots[i]) == slot_id)
		    return token->priv->module->slots[i];

    return NULL;
}

static gboolean
sc_security_token_fetch_certificates (ScSecurityToken *token)
{
    PK11SlotInfo *slot;
    CERTCertList *certificates;
    CERTCertListNode *node;
    SECStatus status;
    int i;

    sc_security_token_unlock (token, "0000");

    sc_debug ("fetching certificates for token in slot %lu",
	      token->priv->slot_id);

    slot = sc_security_token_find_slot_from_id (token,
						token->priv->slot_id);

    g_assert (PK11_GetSlotID (slot) == token->priv->slot_id);

    if (i == token->priv->module->slotCount) {
	    sc_debug ("could not find slot %lu", token->priv->slot_id);
	    return FALSE;
    }

    certificates = PK11_ListCertsInSlot (slot);

    sc_debug ("filtering out non-user certificates");
    if (CERT_FilterCertListForUserCerts (certificates) != SECSuccess) {
	    CERT_DestroyCertList (certificates);
	    sc_debug ("could not filter out non-user certificates");
	    return FALSE;
    }

    for (node = CERT_LIST_HEAD (certificates); 
	 !CERT_LIST_END (node, certificates);
	 node = CERT_LIST_NEXT(node)) {

	    SECCertificateUsage cert_usages;

	    sc_debug ("verifying certificate for use");
	    status = CERT_VerifyCertificateNow (NULL, node->cert, TRUE, 
						0, NULL, &cert_usages);

	    if (status != SECSuccess) {
		    sc_debug ("could not be verified, skipping...");
		    continue;
	    }

	    sc_debug ("got cert with usages 0x%lx", (gulong) cert_usages);

	    if (token->priv->encryption_certificate == NULL) {

		    sc_debug ("checking if certificate can be used for data "
			      "encryption");
		    status = CERT_CheckCertUsage (node->cert, 
						  KU_DATA_ENCIPHERMENT);

		    if (status == SECSuccess) {
			    token->priv->encryption_certificate = 
				    CERT_DupCertificate (node->cert);
		    } else {
			    sc_debug ("certificate can not be used for encryption");
		    }
	    }

	    if (token->priv->signing_certificate == NULL) {

		    sc_debug ("checking if certificate can be used for data "
			      "signing");
		    status = CERT_CheckCertUsage (node->cert, 
						  KU_DIGITAL_SIGNATURE);

		    if (status == SECSuccess) {
			    token->priv->signing_certificate = 
				    CERT_DupCertificate (node->cert);
		    } else {
			    sc_debug ("certificate can not be used for signing things");
		    }
	    }
    }
    return TRUE;
}

#ifdef SC_SECURITY_TOKEN_ENABLE_TEST
#include <glib.h>

static GMainLoop *event_loop;

int 
main (int   argc, 
      char *argv[])
{
    ScSecurityToken *token;
    GError *error;

    g_log_set_always_fatal (G_LOG_LEVEL_ERROR
			    | G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING);

    g_type_init ();

    g_message ("creating instance of 'security token' object...");
    token = _sc_security_token_new (NULL, 1, 1);
    g_message ("'security token' object created successfully");

    g_message ("destroying previously created 'security token' object...");
    g_object_unref (token);
    token = NULL;
    g_message ("'security token' object destroyed successfully");

    return 0;
}
#endif
